import gzip
import inspect
from multiprocessing import Process
from typing import Callable

import typer
import uvicorn
from fastapi import FastAPI, Request, Response
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.routing import APIRoute
from fastapi_pagination import add_pagination
from starlette.middleware.cors import CORSMiddleware

from ._utils import singleton
from .middlewares import process_middleware
from .utils import sync_exec


class GzipRequest(Request):
    async def body(self) -> bytes:
        if not hasattr(self, "_body"):
            body = await super().body()
            if "gzip" in self.headers.getlist("Content-Encoding"):
                body = gzip.decompress(body)
            self._body = body
        return self._body


class GzipRoute(APIRoute):
    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: Request) -> Response:
            request = GzipRequest(request.scope, request.receive)
            return await original_route_handler(request)

        return custom_route_handler


@singleton
class Bali:
    def __init__(self, base_settings=None, **kwargs):
        self.base_settings = base_settings or {}
        self.kwargs = kwargs
        self._app = None

    def __getattribute__(self, attr, *args, **kwargs):
        try:
            return super().__getattribute__(attr)
        except AttributeError:
            if not self._app:
                print(f'attr: {attr}')
                # uvicorn entry is __call__
                if attr == '__call__' or attr == '__getstate__':
                    self.http()
                    return getattr(self._app, attr)

                raise Exception('FastAPI App not initialized')

            return getattr(self._app, attr)

    async def __call__(self, *args, **kwargs):
        self.http()
        await self._app.__call__(*args, **kwargs)

    def _launch_http(self):
        self._app = FastAPI(**self.base_settings)
        uvicorn.run(
            "main:app",
            host="0.0.0.0",
            port=8000,
            reload=True,
            access_log=True
        )

    async def _launch_rpc(self):
        service = self.kwargs.get('rpc_service')
        if not service:
            raise Exception('rpc_service not provided')
        if inspect.iscoroutinefunction(service.serve):
            await service.serve()
        else:
            service.serve()

    def _launch_event(self):
        from .events import handle
        while True:
            handle()

    def _start_all(self):
        process_http = Process(target=self._launch_http)
        process_http.start()
        process_rpc = Process(target=self._launch_rpc)
        process_rpc.start()
        process_rpc.join()
        process_http.join()

    def settings(self, **kwargs):
        self.base_settings.update(kwargs)

    def http(self):
        """load FastAPI to instance"""
        self._app = FastAPI(**self.base_settings)
        self._app.router.route_class = GzipRoute

        # routers
        for router in self.kwargs.get('routers', []):
            self._app.include_router(**router)
        # cors
        backend_cors_origins = self.kwargs.get('backend_cors_origins')
        if backend_cors_origins:
            self._app.add_middleware(GZipMiddleware)
            self._app.add_middleware(
                CORSMiddleware,
                allow_origins=[str(origin) for origin in backend_cors_origins],
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )
            self._app.middleware('http')(process_middleware)

        add_pagination(self._app)

    def launch(
        self, http: bool = False, rpc: bool = False, event: bool = False
    ):
        if not http and not rpc and not event:
            typer.echo(
                'Please provided launch service type: --http or --rpc or --event'
            )

        if http:
            self._launch_http()

        if rpc:
            sync_exec(self._launch_rpc())

        if event:
            self._launch_event()

    def start(self):
        typer.run(self.launch)



import socket
from itertools import groupby

from kombu import Connection, Queue, connections, Exchange

from ..core import _settings

REGISTER_EVENT_CALLBACKS = []


class Callback:
    def __init__(self, queue, callback, connection):
        self._queue = queue
        self._callback = callback
        self._connection = connection

    def __str__(self):
        return f'{self._connection}/{self._queue.name}'

    @property
    def connection(self):
        return self._connection

    @property
    def queue(self):
        return self._queue

    @property
    def callback(self):
        return self._callback


def register_callback(event_type, callback):
    amqp_config_keys = _settings.EVENT_TYPE_TO_AMQP.get(event_type, 'default')
    if not amqp_config_keys:
        raise Exception(
            'Can not find key:%s at EVENT_TYPE_TO_AMQP' % event_type
        )
    arr = amqp_config_keys.split(',')
    for amqp_config_key in arr:
        amqp_config = _settings.AMQP_CONFIGS.get(amqp_config_key)
        if not amqp_config:
            raise Exception(
                'Can not find key:%s at AMQP_CONFIGS' % amqp_config_key
            )
        routing_key = None
        exchange_type = amqp_config.get('EXCHANGE_TYPE')
        if exchange_type != 'fanout':
            routing_key = amqp_config.get(
                'ROUTING_KEY'
            ) or _settings.BALI_ROUTING_KEY.format(event_type)
        exchange = Exchange(
            amqp_config.get('EXCHANGE_NAME', _settings.BALI_EXCHANGE),
            type=exchange_type
        )
        queue = Queue(
            amqp_config.get('QUEUE_NAME') or
            _settings.BALI_QUEUE.format(event_type),
            exchange=exchange,
            routing_key=routing_key
        )
        global REGISTER_EVENT_CALLBACKS
        REGISTER_EVENT_CALLBACKS.append(
            Callback(queue, callback, amqp_config['AMQP_SERVER_ADDRESS'])
        )


def get_connection(amqp_address):
    c1 = Connection(amqp_address)
    return connections[c1].acquire(block=True)


def handle():
    groups = groupby(REGISTER_EVENT_CALLBACKS, key=lambda x: str(x))
    for k, items in groups:
        items = list(items)
        with get_connection(amqp_address=items[0].connection) as conn:
            with conn.Consumer(
                queues=[items[0].queue],
                accept=['json'],
                callbacks=[i.callback for i in items],
                auto_declare=False
            ) as consumer:
                try:
                    conn.drain_events(timeout=2)
                except socket.timeout:
                    pass
                except Exception as e:
                    raise e
    return True



import logging
import warnings
from functools import wraps

from sqla_wrapper import SQLAlchemy, BaseModel
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.decl_api import DeclarativeMeta

from .models import included_models

# TODO: Removed logging according 12factor
error_logger = logging.getLogger('error')

# SQLA-Wrapper 4.x supported session proxy methods
# https://github.com/jpsca/sqla-wrapper/blob/v4.200628/sqla_wrapper/session_proxy.py
SQLA_WRAPPER_SESSION_PROXIES = (
    'query',
    'add',
    'add_all',
    'begin',
    'begin_nested',
    'commit',
    'delete',
    'execute',
    'expire',
    'expire_all',
    'expunge',
    'expunge_all',
    'flush',
    'invalidate',
    'is_modified',
    'merge',
    'prepare',
    'prune',
    'refresh',
    'remove',
    'rollback',
    'scalar',
)


# noinspection PyPep8Naming
class DB:
    def __init__(self):
        self._db = None  # sync engine
        self._async_engine = None  # async engine
        self.async_session = None  # async session maker

    def connect(
        self,
        database_uri,
        engine_options=None,
        session_options=None,
        **kwargs
    ):
        engine_options = engine_options or {}
        engine_options.setdefault("pool_size", 5)
        engine_options.setdefault("pool_recycle", 2 * 60 * 60)

        session_options = session_options or {}
        # developers need to know when the ORM object needs to reload
        # from the db
        session_options.setdefault("expire_on_commit", False)

        # Sync mode db instance
        self._db = SQLAlchemy(
            database_uri,
            engine_options=engine_options,
            session_options=session_options,
        )
        self._db.Model = self._db.registry.generate_base(
            cls=BaseModel,
            name="Model",
            metaclass=AsyncModelDeclarativeMeta,
        )

        async_database_uri = get_async_database_uri(database_uri)
        self._async_engine = create_async_engine(async_database_uri)

        self.async_session = sessionmaker(
            self._async_engine,
            class_=AsyncSession,
            expire_on_commit=False,
        )

    def __getattribute__(self, attr, *args, **kwargs):
        try:
            return super().__getattribute__(attr)
        except AttributeError:
            if not self._db:
                raise Exception('Database session not initialized')

            # BaseModels
            if attr in included_models:
                return included_models[attr](self)

            # Compatible legacy SQLA-Wrapper SessionProxy
            if attr in SQLA_WRAPPER_SESSION_PROXIES:
                return getattr(self._db.s, attr)

            return getattr(self._db, attr)


db = DB()


def get_async_database_uri(database_uri):
    """
    Transform populate database schema to async format,
    which is used by SQLA-Wrapper
    """
    uri = database_uri
    database_schema_async_maps = [
        ('sqlite://', 'sqlite+aiosqlite://'),
        ('mysql+pymysql://', 'mysql+aiomysql://'),
        ('postgres://', 'postgresql+asyncpg://'),
    ]
    for sync_schema, async_schema in database_schema_async_maps:
        uri = uri.replace(sync_schema, async_schema)
    return uri


MAXIMUM_RETRY_ON_DEADLOCK: int = 3


def retry_on_deadlock_decorator(func):
    warnings.warn(
        'retry_on_deadlock_decorator will remove in 3.2',
        DeprecationWarning,
    )

    lock_messages_error = [
        'Deadlock found',
        'Lock wait timeout exceeded',
    ]

    @wraps(func)
    def wrapper(*args, **kwargs):
        attempt_count = 0
        while attempt_count < MAXIMUM_RETRY_ON_DEADLOCK:
            try:
                return func(*args, **kwargs)
            except OperationalError as e:
                # noinspection PyUnresolvedReferences
                if any(msg in e.message for msg in lock_messages_error) \
                        and attempt_count <= MAXIMUM_RETRY_ON_DEADLOCK:
                    error_logger.error(
                        'Deadlock detected. Trying sql transaction once more. '
                        'Attempts count: %s' % (attempt_count + 1)
                    )
                else:
                    raise
            attempt_count += 1

    return wrapper


def close_connection(func):
    warnings.warn(
        'retry_on_deadlock_decorator will remove in 3.2',
        DeprecationWarning,
    )

    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        finally:
            db.remove()

        return result

    return wrapper


class AsyncModelDeclarativeMeta(DeclarativeMeta):
    """Make db.BaseModel support async using this metaclass"""
    def __getattribute__(self, attr):
        if attr == 'aio':
            aio = super().__getattribute__(attr)
            if any([aio.db is None, aio.model is None]):
                aio = type(
                    f'Aio{aio.__qualname__}',
                    aio.__bases__,
                    dict(aio.__dict__),
                )
                setattr(aio, 'db', self._db)
                setattr(aio, 'model', self)
            return aio

        return super().__getattribute__(attr)

    # noinspection PyMethodParameters
    def __call__(self, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)
        instance.aio = self.aio(instance)
        return instance



import logging.config
from typing import Dict

from pydantic import BaseSettings

# noinspection PyUnresolvedReferences
from .application import Bali
from .cache import cache
from .cache_memoize import cache_memoize
from .db import db, transaction
# noinspection PyUnresolvedReferences
from .routing import APIRouter

# Bind transaction to db
setattr(db, 'transaction', transaction)


class Settings(BaseSettings):
    ENABLED_RPC_LOGGING: str = False
    AMQP_CONFIGS: Dict[str, Dict[str, str]] = {
        'default':
            {
                'AMQP_SERVER_ADDRESS': 'amqp://user:passwd@host:port',
                # It will be used if you defined EXCHANGE_NAME, else
                # BALI_EXCHANGE will be used
                # 'EXCHANGE_NAME': 'event_default_exchange',
                # It will be used if you set queue name,else queue name
                # will be generated by BALI_QUEUE and event_type
                # 'QUEUE_NAME': 'event_default_queue',
                # It will used if you set routing_key,else it
                # will be generated by BALI_ROUTING_KEY and
                # event_type
                # 'ROUTING_KEY': 'event_default_routing_key'
            }
    }
    # You can map event_type to amqp, or it will be map to default amqp
    # You can map event_type to more than one amqp split by ','
    # EVENT_TYPE_TO_AMQP = {'test0': 'default', 'test1': 'default,others'}
    EVENT_TYPE_TO_AMQP: Dict[str, str] = {}
    BALI_QUEUE = 'event_default_queue'
    BALI_EXCHANGE = 'event_default_exchange'
    BALI_ROUTING_KEY = 'event_default_routing_key'


_settings = Settings()


def initialize(settings):
    # update basic settings
    if not hasattr(settings, 'ENABLED_RPC_LOGGING'):
        settings.Config.__setattr__(settings, 'ENABLED_RPC_LOGGING', False)

    # load custom settings
    global _settings
    _settings = settings

    # initialize db connections, default enabled db
    # can be disabled by `DISABLE_DB_CONNECTION` settings
    if not getattr(settings, 'DISABLE_DB_CONNECTION', False):
        if not hasattr(settings, 'SQLALCHEMY_DATABASE_URI'):
            raise Exception(
                'Initialized db connection without `SQLALCHEMY_DATABASE_URI` setting'
            )

        db.connect(settings.SQLALCHEMY_DATABASE_URI)

    # initialize cache connections, default enabled db
    # can be disabled by `DISABLED_DB_CONNECTION` settings
    # cache prefix can be custom by `CACHE_PREFIX`
    if not getattr(settings, 'DISABLE_CACHE_CONNECTION', False):
        if not hasattr(settings, 'CACHE_ADDRESS'):
            raise Exception(
                'Initialized cache connection without `CACHE_ADDRESS` setting'
            )

        if not hasattr(settings, 'CACHE_PASSWORD'):
            raise Exception(
                'Initialized cache connection without `CACHE_PASSWORD` setting'
            )

        cache.connect(
            settings.CACHE_ADDRESS,
            password=settings.CACHE_PASSWORD,
            prefix=getattr(
                settings, 'CACHE_PREFIX', f'{settings.SERVER_NAME}_service'
            )
        )
    if hasattr(settings, 'LOGGING_CONFIG'):
        logging.config.dictConfig(settings.LOGGING_CONFIG)

    if not hasattr(settings, 'BALI_QUEUE'):
        settings.Config.__setattr__(settings, 'BALI_QUEUE', '{}.events')

    if not hasattr(settings, 'BALI_EXCHANGE'):
        settings.Config.__setattr__(settings, 'BALI_EXCHANGE', 'ms.events')

    if not hasattr(settings, 'BALI_ROUTING_KEY'):
        settings.Config.__setattr__(settings, 'BALI_ROUTING_KEY', '{}.routing')
    if not hasattr(settings, 'EVENT_TYPE_TO_AMQP'):
        settings.Config.__setattr__(settings, 'EVENT_TYPE_TO_AMQP', {})



"""Generic Routes

Generic routes include list/get/create/update/delete
Used by `resource.RouterGenerator`

# TODO: Optimized boilerplate code

"""

import inspect
import logging
import typing
from typing import Callable

from fastapi import Request

from ..paginate import paginate
from ..schemas import ListRequest


# --- Utilities functions for route handler ---
def pick_route(func, async_route, route):
    """Pick coroutine or generic function for route"""
    return async_route if inspect.iscoroutinefunction(func) else route


# noinspection PyUnresolvedReferences,PyProtectedMember
def list_(router_generator) -> Callable:
    """
    list action default using fastapi-pagination to process paginate
    """
    resource = router_generator.cls()
    action_func = getattr(resource, 'list')

    # noinspection PyProtectedMember,PyShadowingNames,PyUnresolvedReferences
    def route(request: Request = None):
        resource._request = request
        router_generator.check_permissions(resource)
        params = request.query_params._dict

        # parse filters
        filters = {}
        for k, v in params.items():
            if k not in router_generator._ordered_filters:
                continue
            # noinspection PyBroadException
            try:
                # Convert param and retrieve param
                params_converter = router_generator._ordered_filters.get(k)
                if isinstance(params_converter, typing._GenericAlias):
                    params_converter = params_converter.__args__[0]
                filters[k] = params_converter(v)
            except Exception as ex:
                logging.warning(
                    'Query params `%s`(value: %s) type convert failed, '
                    'exception: %s', k, v, ex
                )
                continue

        schema_in = ListRequest(**params, filters=filters)

        result = action_func(schema_in)
        return paginate(result)

    # noinspection PyProtectedMember,PyShadowingNames,PyUnresolvedReferences
    async def async_route(request: Request = None):
        resource._request = request
        router_generator.check_permissions(resource)
        params = request.query_params._dict

        # parse filters
        filters = {}
        for k, v in params.items():
            if k not in router_generator._ordered_filters:
                continue
            # noinspection PyBroadException
            try:
                # Convert param and retrieve param
                params_converter = router_generator._ordered_filters.get(k)
                if isinstance(params_converter, typing._GenericAlias):
                    params_converter = params_converter.__args__[0]
                filters[k] = params_converter(v)
            except Exception as ex:
                logging.warning(
                    'Query params `%s`(value: %s) type convert failed, '
                    'exception: %s', k, v, ex
                )
                continue

        schema_in = ListRequest(**params, filters=filters)

        result = await action_func(schema_in)
        return paginate(result)

    # update route's signatures
    parameters = []
    for k, v in router_generator._ordered_filters.items():
        default = inspect.Parameter.empty
        if isinstance(v, typing._GenericAlias):
            default = None if type(None) in v.__args__ else default
        parameters.append(
            inspect.Parameter(
                name=k,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=default,
                annotation=v,
            )
        )
    sig = inspect.signature(route)
    parameters.extend(sig.parameters.values())
    route.__signature__ = sig.replace(parameters=parameters)

    return pick_route(action_func, async_route, route)


def get_(router_generator) -> Callable:
    resource = router_generator.cls()
    action_func = getattr(resource, 'get')

    def route(id: int, request: Request = None):
        resource._request = request
        router_generator.check_permissions(resource)
        return action_func(pk=id)

    async def async_route(id: int, request: Request = None):
        resource._request = request
        router_generator.check_permissions(resource)
        return await action_func(pk=id)

    return pick_route(action_func, async_route, route)


def create_(router_generator) -> Callable:
    resource = router_generator.cls()
    action_func = getattr(resource, 'create')

    def route(schema_in: resource.schema, request: Request = None):
        resource._request = request
        router_generator.check_permissions(resource)
        return action_func(schema_in)

    async def async_route(schema_in: resource.schema, request: Request = None):
        resource._request = request
        router_generator.check_permissions(resource)
        return await action_func(schema_in)

    return pick_route(action_func, async_route, route)


def update_(router_generator) -> Callable:
    resource = router_generator.cls()
    action_func = getattr(resource, 'update')

    def route(schema_in: resource.schema, id: int, request: Request = None):
        resource._request = request
        router_generator.check_permissions(resource)
        return action_func(schema_in, pk=id)

    async def async_route(
        schema_in: resource.schema, id: int, request: Request = None
    ):
        resource._request = request
        router_generator.check_permissions(resource)
        return await action_func(schema_in, pk=id)

    return pick_route(action_func, async_route, route)


def delete_(router_generator) -> Callable:
    resource = router_generator.cls()
    action_func = getattr(resource, 'delete')

    def route(id: int, request: Request = None):
        resource._request = request
        router_generator.check_permissions(resource)
        return action_func(pk=id)

    async def async_route(id: int, request: Request = None):
        resource._request = request
        router_generator.check_permissions(resource)
        return await action_func(pk=id)

    return pick_route(action_func, async_route, route)



"""Model included

`db.BaseModel` is the most common model.
If you don't use `db.BaseModel`, you can compose Mixins to `db.Model`

Import Mixins in your project examples:

    ```python
    from bali.db.models import GenericModelMixin, AsyncModelMixin
    ```

"""

from contextvars import ContextVar
from datetime import datetime
from typing import List, Dict

import pytz
from sqlalchemy import Column, DateTime, Boolean
from sqlalchemy.exc import IntegrityError, SQLAlchemyError
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.future import select
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.functions import func
from sqlalchemy.types import TypeDecorator

from ..utils import timezone


class AwareDateTime(TypeDecorator):
    impl = DateTime

    @property
    def python_type(self):
        return datetime

    def process_result_value(self, value, _):
        if value is not None:
            value = timezone.make_aware(value, timezone=pytz.utc)

        return value

    def process_bind_param(self, value, _):
        if value is not None and timezone.is_aware(value):
            value = timezone.make_naive(value, timezone=pytz.utc)

        return value


context_auto_commit = ContextVar('context_auto_commit', default=True)


def get_base_model(db):
    class BaseModel(db.Model, GenericModelMixin, AsyncModelMixin):
        __abstract__ = True
        __asdict_include_hybrid_properties__ = False

        # Bind SQLA-wrapper database to model
        _db = db

        @classmethod
        def exists(cls, **attrs):
            """Returns whether an object with these attributes exists."""
            query = cls.query().filter_by(**attrs).exists()
            return bool(db.s.query(query).scalar())

        @classmethod
        def create(cls, **attrs):
            """
            Create and persist a new record for the model, and returns it.
            """
            return cls(**attrs).save()

        @classmethod
        def create_or_first(cls, **attrs):
            """Tries to create a new record, and if it fails
            because already exists, return the first it founds."""
            try:
                return cls.create(**attrs)
            except IntegrityError:
                db.s.rollback()
                return cls.first(**attrs)

        @classmethod
        def first(cls, **attrs):
            """Returns the first object found with these attributes."""
            return cls.query().filter_by(**attrs).first()

        @classmethod
        def first_or_error(cls, **attrs):
            """Returns the first object found with these attributes
            or raises a `ValuError` if it doesn't find one."""
            obj = cls.first(**attrs)
            if obj is None:
                raise ValueError
            return obj

        @classmethod
        def query(cls):
            return db.s.query(cls)

        def save(self):
            """Override default model's save"""
            global context_auto_commit
            db.s.add(self)
            db.s.commit() if context_auto_commit.get() else db.s.flush()
            return self

        def delete(self):
            """Override default model's delete"""
            global context_auto_commit
            db.s.delete(self)
            db.s.commit() if context_auto_commit.get() else db.s.flush()

        def _asdict(self, **kwargs):
            include_hybrid_properties = kwargs.setdefault(
                "include_hybrid_properties",
                self.__asdict_include_hybrid_properties__
            )

            output_fields = []
            for i in inspect(type(self)).all_orm_descriptors:
                if isinstance(i, InstrumentedAttribute):
                    output_fields.append(i.key)
                elif isinstance(
                    i, hybrid_property
                ) and include_hybrid_properties:
                    output_fields.append(i.__name__)

            return {i: getattr(self, i, None) for i in output_fields}

        dict = to_dict = _asdict

        @classmethod
        def count(cls, **attrs) -> int:
            return db.s.query(func.count(cls.id)).filter_by(**attrs).scalar()

        @classmethod
        def get_fields(cls) -> List[str]:
            return [c.name for c in cls.__table__.columns]

        @classmethod
        def get_or_create(cls, defaults: Dict = None, **kwargs):
            instance = db.s.query(cls).filter_by(**kwargs).one_or_none()
            if instance:
                return instance, False

            instance = cls(**{**kwargs, **(defaults or {})})  # noqa
            try:
                db.s.add(instance)
                db.s.commit()
                return instance, True
            except SQLAlchemyError:
                db.s.rollback()
                instance = db.s.query(cls).filter_by(**kwargs).one()
                return instance, False

        @classmethod
        def update_or_create(cls, defaults: Dict = None, **kwargs):
            try:
                try:
                    instance = (
                        db.s.query(cls).filter_by(
                            **kwargs
                        ).populate_existing().with_for_update().one()
                    )
                except NoResultFound:
                    instance = cls(**{**kwargs, **(defaults or {})})  # noqa
                    try:
                        db.s.add(instance)
                        db.s.commit()
                    except SQLAlchemyError:
                        db.s.rollback()
                        instance = (
                            db.s.query(cls).filter_by(
                                **kwargs
                            ).populate_existing().with_for_update().one()
                        )
                    else:
                        return instance, True
            except SQLAlchemyError:
                db.s.rollback()
                raise
            else:
                for k, v in defaults.items():
                    setattr(instance, k, v)
                db.s.add(instance)
                db.s.commit()
                db.s.refresh(instance)
                return instance, False

    return BaseModel


class AsyncModelManager:
    """Async model bind to aio"""

    db = None
    model = None

    def __init__(self, instance):
        self.instance = instance

    @classmethod
    async def exists(cls, **attrs):
        async with cls.db.async_session() as async_session:
            stmt = select(cls.model).filter_by(**attrs)
            result = await async_session.execute(stmt)
            return bool(result.scalars().first())

    @classmethod
    async def create(cls, **attrs):
        await cls.model(**attrs).aio.save()

    @classmethod
    async def first(cls, **attrs):
        async with cls.db.async_session() as async_session:
            stmt = select(cls.model).filter_by(**attrs)
            result = await async_session.execute(stmt)
            return result.scalars().first()

    async def save(self):
        async with self.db.async_session() as async_session:
            async_session.add(self.instance)
            await async_session.commit()
            return self.instance

    async def delete(self):
        async with self.db.async_session() as async_session:
            async_session.delete(self.instance)
            await async_session.commit()


# expose the include models and model creator
included_models = {
    'BaseModel': get_base_model,
}

# --------------------  Models Mixins  -------------------- #


class GenericModelMixin:
    """Generic model include the following fields"""
    created_time = Column(DateTime, default=datetime.utcnow)
    updated_time = Column(
        DateTime, default=datetime.utcnow, onupdate=datetime.utcnow
    )
    is_active = Column(Boolean, default=True)


class AsyncModelMixin:
    """Async models methods

    All async method accessed by `aio`

        ```python
        # Model
        async def get_first_user()
            await User.aio.first()

        # Instance
        user = User()
        await user.aio.save()
        ```

    """

    aio = AsyncModelManager



from kombu import Connection, Queue, Exchange

from .event import Event
from ..core import _settings


def dispatch(event: Event, amqp_name: str = ''):
    if not amqp_name:
        amqp_name = event.amqp_name
    amqp_config = _settings.AMQP_CONFIGS.get(amqp_name, {})
    if not amqp_config:
        raise Exception('Settings not set:sub config of AMQP_CONFIGS not set')
    if 'AMQP_SERVER_ADDRESS' not in amqp_config:
        raise Exception('Settings not set: AMQP_SERVER_ADDRESS of sub config')

    routing_key = None
    exchange_type = amqp_config.get('EXCHANGE_TYPE')
    exchange = Exchange(
        name=amqp_config.get('EXCHANGE_NAME') or _settings.BALI_EXCHANGE,
        type=exchange_type
    )
    declare = [exchange]
    if exchange_type != 'fanout':
        routing_key = amqp_config.get(
            'ROUTING_KEY'
        ) or _settings.BALI_ROUTING_KEY.format(event.type)
    if amqp_config.get('QUEUE_NAME'):
        queue = Queue(
            amqp_config.get('QUEUE_NAME'),
            exchange,
            routing_key=routing_key
        )
        declare.append(queue)
    with Connection(amqp_config['AMQP_SERVER_ADDRESS']) as conn:
        # produce
        producer = conn.Producer(serializer='json')
        return producer.publish(
            event.dict(),
            exchange=exchange,
            routing_key=routing_key,
            declare=declare,
            retry=True
        )



from collections import defaultdict

from sqlalchemy import case
from sqlalchemy.ext.hybrid import Comparator
from sqlalchemy.util.langhelpers import dictlike_iteritems


class CaseComparator(Comparator):
    def __init__(self, whens, expression):
        super().__init__(expression)
        self.whens, self.reversed_whens = dictlike_iteritems(whens), defaultdict(list)
        for k, v in self.whens:
            self.reversed_whens[v].append(k)

    def __clause_element__(self):
        return case(self.whens, self.expression)

    def __eq__(self, other):
        return super().__clause_element__().in_(self.reversed_whens[other])
